import { Component, Prop, h, Watch, Host, Listen, Event, State, } from "@stencil/core";
import isEqual from 'lodash/isEqual';
import gql from 'graphql-tag';
import { ApolloClient, InMemoryCache } from '@apollo/client';
import { GraphQLWsLink } from '@apollo/client/link/subscriptions';
import { createClient } from 'graphql-ws';
import WebSocket from 'ws';
import ResponsesService from '../../aggregates/responses-service';
import { loadAggregator } from '../../response-loader/response-loader';
import { loadAssignmentItems, loadAssignmentSessions } from '../../load-config';
const environmentVariables = {
    dev: process.env.DEV,
    prod: process.env.PROD,
    apiDev: process.env.API_DEV,
    apiProd: process.env.API_PROD,
};
const watchAssignmentSessionsQuery = gql `
    subscription watchAssignmentSessions($assignmentId: ID!){
      watchAssignmentSessions(assignmentId: $assignmentId) {
          id
          item {
            id
            name
            config
            version {
            major
            minor
            patch
            prerelease
          }
        }
        mode
        data
        autoScore
        manualScore,
        createdAt
      }
    }
`;
const watchAssignmentItemsQuery = gql `
  subscription watchAssignmentItems($assignmentId: ID!){
    watchAssignmentItems(assignmentId: $assignmentId) {
      id
      name
      config
      searchMetaData
      version {
        major
        minor
        patch
        prerelease
      }
    }
  }
`;
export class PieApiLive {
    constructor() {
        this.pageSize = 250;
        /**
         * If loading indicator should show up while the item is loading.
         */
        this.loadingIndicator = true;
        /**
         *  Used to store targets to all pie-live-sessions in order to be able to pass proper data
         */
        this._mapSessionIdsToSessionTargets = {};
        /** Used to store all the items that were received via watchAssignmentItems */
        this._items = {};
        /** Used to store all the sessions that were received via watchAssignmentSessions */
        this._sessions = {};
        /** Used to map the sessions with proper items */
        this._mapSessionIdsToItemIds = {};
        this.setItemAndSession = (target, session, item) => {
            target._itemAndSessionData = {
                item,
                session
            };
        };
        this.setItemAndSessionData = (sessionId) => {
            const target = this._mapSessionIdsToSessionTargets[sessionId];
            const session = sessionId && this._sessions[sessionId];
            const item = session && session.item && session.item.id && this._items[session.item.id];
            if (target && item && session) {
                this.setItemAndSession(target, session, item);
            }
        };
        this.getAssignmentSessionsAndItems = async () => {
            // default is production
            let apiUri = `https://${environmentVariables.apiProd}`;
            // if host property is dev, use the value associated to the dev api for fetching data
            if (this.host && this.host === 'dev') {
                apiUri = `https://${environmentVariables.apiDev}`;
            }
            await this.loadAllAssignmentItems(this.token, this.assignmentId, { 'limit': this.pageSize }, apiUri);
            await this.loadAllAssignmentSessions(this.token, this.assignmentId, { 'limit': this.pageSize }, apiUri);
        };
        this.parseSession = (session) => {
            const { id, item } = session;
            this._mapSessionIdsToItemIds[id] = item.id;
            if (!this._sessions[id]) {
                this.newSessionInAssignment.emit({ session });
            }
            this._sessions[id] = Object.assign(Object.assign({}, session), { item });
            this.setItemAndSessionData(id);
        };
        this.parseItem = (item) => {
            const { id } = item;
            if (!this._items[id]) {
                this.newItemInAssignment.emit({ item });
            }
            this._items[id] = item;
            Object.keys(this._mapSessionIdsToItemIds)
                .filter(key => this._mapSessionIdsToItemIds[key] === id)
                .forEach(this.setItemAndSessionData);
        };
        this.loadAllAssignmentItems = async (token, assignmentId, itemOpts, apiUrl) => {
            var _a, _b, _c;
            let allItems = [];
            let cursorNext = null;
            do {
                try {
                    const result = await loadAssignmentItems(token, assignmentId, itemOpts, apiUrl);
                    if (!result)
                        break;
                    const items = ((_a = result === null || result === void 0 ? void 0 : result.items) === null || _a === void 0 ? void 0 : _a.docs) || [];
                    allItems = [...allItems, ...items];
                    items.forEach((item) => this.parseItem(item));
                    cursorNext = (_c = (_b = result === null || result === void 0 ? void 0 : result.items) === null || _b === void 0 ? void 0 : _b.cursor) === null || _c === void 0 ? void 0 : _c.next;
                    itemOpts = Object.assign(Object.assign({}, itemOpts), { cursor: cursorNext });
                }
                catch (e) {
                    break;
                }
            } while (cursorNext);
            return allItems;
        };
        this.loadAllAssignmentSessions = async (token, assignmentId, sessionOpts, apiUrl) => {
            var _a, _b, _c;
            let allSessions = [];
            let cursorNext = null;
            do {
                try {
                    const result = await loadAssignmentSessions(token, assignmentId, sessionOpts, apiUrl);
                    if (!result)
                        break;
                    const sessions = ((_a = result === null || result === void 0 ? void 0 : result.sessions) === null || _a === void 0 ? void 0 : _a.docs) || [];
                    allSessions = [...allSessions, ...sessions];
                    sessions.forEach((session) => this.parseSession(session));
                    cursorNext = (_c = (_b = result === null || result === void 0 ? void 0 : result.sessions) === null || _b === void 0 ? void 0 : _b.cursor) === null || _c === void 0 ? void 0 : _c.next;
                    sessionOpts = Object.assign(Object.assign({}, sessionOpts), { cursor: cursorNext });
                }
                catch (e) {
                    break;
                }
            } while (cursorNext);
            return allSessions;
        };
        this.subscribeAssignmentSessions = () => {
            if (!this._subscriptionForSessions) {
                this._subscriptionForSessions = this.gqlClient.subscribe({
                    query: watchAssignmentSessionsQuery,
                    variables: { assignmentId: this.assignmentId },
                })
                    .subscribe(result => {
                    var _a;
                    const session = (_a = result === null || result === void 0 ? void 0 : result.data) === null || _a === void 0 ? void 0 : _a.watchAssignmentSessions;
                    this.parseSession(session);
                }, error => {
                    console.error(error.message);
                });
            }
        };
        this.subscribeAssignmentItems = () => {
            if (!this._subscriptionForItems) {
                this._subscriptionForItems = this.gqlClient.subscribe({
                    query: watchAssignmentItemsQuery,
                    variables: { assignmentId: this.assignmentId },
                }).subscribe(result => {
                    var _a;
                    const item = (_a = result === null || result === void 0 ? void 0 : result.data) === null || _a === void 0 ? void 0 : _a.watchAssignmentItems;
                    this.parseItem(item);
                }, error => {
                    console.error(error.message);
                });
            }
        };
        this.loadAssignment = () => {
            if (this.gqlClient && this.assignmentId) {
                (async () => {
                    await this.getAssignmentSessionsAndItems();
                    this.subscribeAssignmentItems();
                    this.subscribeAssignmentSessions();
                })();
            }
        };
        this.addSessionId = (event) => {
            const { detail, target } = event || {};
            const { sessionId } = detail || {};
            if (sessionId && target) {
                this._mapSessionIdsToSessionTargets[sessionId] = target;
                this.setItemAndSessionData(sessionId);
            }
        };
    }
    watchChange() {
        this.initClient();
    }
    watchMetadata(newVal, oldVal) {
        if (newVal && !isEqual(newVal, oldVal)) {
            this.loadAssignment();
        }
    }
    watchAssignmentId(newVal, oldVal) {
        if (newVal !== oldVal && newVal) {
            this.assignmentId = newVal;
            this.loadAssignment();
        }
    }
    connectedCallback() {
        this.initClient();
    }
    async disconnectedCallback() {
        var _a, _b, _c;
        if (this.gqlClient) {
            try {
                await ((_a = this.gqlClient) === null || _a === void 0 ? void 0 : _a.stop());
                (_c = (_b = this.gqlClient.link) === null || _b === void 0 ? void 0 : _b.client) === null || _c === void 0 ? void 0 : _c.dispose();
                this.gqlClient = null;
            }
            catch (e) {
                console.error(e.toString());
            }
        }
    }
    handleStartSubscription(event) {
        if (this.gqlClient) {
            this.addSessionId(event);
        }
        else {
            throw new Error('no client available');
        }
    }
    handleLiveSessionAdded(ev) {
        // do not change the order of the assignations
        // these are important as how they are used in pie-live-session @watch
        ev.target.loadingIndicator = this.loadingIndicator;
        ev.target.metadata = this.metadata;
    }
    async calculateAndEmitAggregate(detail) {
        const { data, score, sessionId, itemConfig, aggregateId } = detail || {};
        const aggregatorLoaded = await loadAggregator(itemConfig);
        if (aggregatorLoaded) {
            // add/update the response in responses service
            this.responsesService.upsertAggregate({ aggregateId, score, sessionId, data });
            // get the aggregate for all the responses that belong to the same question
            const aggregates = this.responsesService.getAggregate(itemConfig, aggregateId);
            // emit the aggregate (it is currently handled in pie-aggregate)
            // todo this only takes first aggregate for now (can be updated to support multi-part in the future)
            this.itemAggregate.emit({ aggregate: aggregates[0], aggregateId, itemConfig });
        }
    }
    async handleUpdateAggregate(event) {
        if (this.emitAggregates && this.responsesService) {
            await this.calculateAndEmitAggregate(event.detail);
        }
    }
    initClient() {
        const isTestMode = process.env.NODE_ENV === 'test';
        // set LIVE SERVER HOST (for subscriptions)
        if (!this.gqlClient && this.token) {
            // default is production
            let liveUri = `wss://${environmentVariables.prod}`;
            // if host property is defined, determine live server using that value
            if (this.host) {
                liveUri = `wss://${environmentVariables[this.host]}`;
            }
            // if live server CUSTOM host property is defined, use that value
            if (this.liveUrl) {
                liveUri = this.liveUrl;
            }
            // set GQL CLIENT
            const wsLink = typeof window !== 'undefined'
                ? new GraphQLWsLink(createClient(Object.assign(Object.assign({ url: liveUri }, (isTestMode && { webSocketImpl: WebSocket })), { connectionParams: {
                        Authorization: `Bearer ${this.token}`
                    } }))) : null;
            this.gqlClient = new ApolloClient({
                ssrMode: typeof window === 'undefined',
                link: wsLink,
                cache: new InMemoryCache(),
            });
            this.loadAssignment();
        }
        if (this.emitAggregates) {
            if (!this.responsesService) {
                this.responsesService = new ResponsesService();
            }
        }
    }
    render() {
        return h(Host, { connected: this.gqlClient ? "true" : "false" });
    }
    static get is() { return "pie-api-live"; }
    static get originalStyleUrls() { return {
        "$": ["pie.css"]
    }; }
    static get styleUrls() { return {
        "$": ["pie.css"]
    }; }
    static get properties() { return {
        "token": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "string",
                "resolved": "string",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "The api token for pie-api.\nThe token must have the `live` scope."
            },
            "attribute": "token",
            "reflect": false
        },
        "host": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "string",
                "resolved": "string",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "The pie-api-live host (environment) - used to determine live (subscriptions) hosts.\nOne of: `dev`, `prod`.\n\ne.g.: if host = 'dev' => webSocket connection to wss://live.dev.pie-api.com/api/graphql"
            },
            "attribute": "host",
            "reflect": false
        },
        "liveUrl": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "string",
                "resolved": "string",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [{
                        "text": undefined,
                        "name": "internal"
                    }],
                "text": "The live server CUSTOM host. If present, overrides the live (subscriptions) server host.\n\ne.g.: `wss://live.dev.pie-api.com/api/graphql"
            },
            "attribute": "live-url",
            "reflect": false
        },
        "pageSize": {
            "type": "number",
            "mutable": false,
            "complexType": {
                "original": "number",
                "resolved": "number",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": ""
            },
            "attribute": "page-size",
            "reflect": false,
            "defaultValue": "250"
        },
        "assignmentId": {
            "type": "string",
            "mutable": false,
            "complexType": {
                "original": "string",
                "resolved": "string",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "The Assignment ID - REQUIRED in order to load any data"
            },
            "attribute": "assignment-id",
            "reflect": false
        },
        "loadingIndicator": {
            "type": "boolean",
            "mutable": false,
            "complexType": {
                "original": "boolean",
                "resolved": "boolean",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "If loading indicator should show up while the item is loading."
            },
            "attribute": "loading-indicator",
            "reflect": false,
            "defaultValue": "true"
        },
        "metadata": {
            "type": "unknown",
            "mutable": false,
            "complexType": {
                "original": "MetadataProperty[]",
                "resolved": "MetadataProperty[]",
                "references": {
                    "MetadataProperty": {
                        "location": "import",
                        "path": "../../types"
                    }
                }
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "An array of MetadataProperties to be extracted from Item and added to `pie-live-session` elements as attributes.\nIf no value is found for the provided key on an item, the attribute will not be set.\nThis property must be set before adding any child `<pie-live-session>` elements.\n\n[\n   {\n     key:\"metadata_key\",\n     attribute: \"attribute-name\"\n   }\n]\n\nThis property can alternatively be set declaritively in html using `<item-metadata key=\"foo\" attribute-name=\"foo-value\">` as a child\nelement of `<pie-api-live>` before any `<pie-live-session>` elements."
            }
        },
        "emitAggregates": {
            "type": "boolean",
            "mutable": false,
            "complexType": {
                "original": "boolean",
                "resolved": "boolean",
                "references": {}
            },
            "required": false,
            "optional": false,
            "docs": {
                "tags": [],
                "text": "Boolean that determines if aggregates have to be calculated and emitted"
            },
            "attribute": "emit-aggregates",
            "reflect": false
        }
    }; }
    static get states() { return {
        "gqlClient": {}
    }; }
    static get events() { return [{
            "method": "itemAggregate",
            "name": "itemAggregate",
            "bubbles": true,
            "cancelable": true,
            "composed": true,
            "docs": {
                "tags": [],
                "text": "Emitted when an aggregate has been recalculated"
            },
            "complexType": {
                "original": "any",
                "resolved": "any",
                "references": {}
            }
        }, {
            "method": "newItemInAssignment",
            "name": "newItemInAssignment",
            "bubbles": true,
            "cancelable": true,
            "composed": true,
            "docs": {
                "tags": [],
                "text": "Emitted when there's a new item in the assignment"
            },
            "complexType": {
                "original": "any",
                "resolved": "any",
                "references": {}
            }
        }, {
            "method": "newSessionInAssignment",
            "name": "newSessionInAssignment",
            "bubbles": true,
            "cancelable": true,
            "composed": true,
            "docs": {
                "tags": [],
                "text": "Emitted when there's a new session in the assignment"
            },
            "complexType": {
                "original": "any",
                "resolved": "any",
                "references": {}
            }
        }]; }
    static get watchers() { return [{
            "propName": "token",
            "methodName": "watchChange"
        }, {
            "propName": "host",
            "methodName": "watchChange"
        }, {
            "propName": "liveUrl",
            "methodName": "watchChange"
        }, {
            "propName": "emitAggregates",
            "methodName": "watchChange"
        }, {
            "propName": "metadata",
            "methodName": "watchMetadata"
        }, {
            "propName": "assignmentId",
            "methodName": "watchAssignmentId"
        }]; }
    static get listeners() { return [{
            "name": "startSubscription",
            "method": "handleStartSubscription",
            "target": "window",
            "capture": false,
            "passive": false
        }, {
            "name": "attachSession",
            "method": "handleLiveSessionAdded",
            "target": undefined,
            "capture": false,
            "passive": false
        }, {
            "name": "updateAggregate",
            "method": "handleUpdateAggregate",
            "target": "window",
            "capture": false,
            "passive": false
        }]; }
}
