const { ApolloServer, gql, PubSub, withFilter, } = require("apollo-server-express");
const { GraphQLJSON } = require("graphql-type-json");
const { makeExecutableSchema } = require("graphql-tools");
const { mockSaveEventData } = require("../test/mock-session-data");
const shortId = require("shortid");
var fs = require("fs");
const pubsub = new PubSub();
const SESSION_EVENT = "SESSION_EVENT";
const typeDefs = gql `
  scalar JSON
  type Query {
    sessionEvent(id: Int!): String!
    contentItemSession(id: ID!): ContentItemSession
  }

  type ContentItemSession {
    id: ID!
    data: JSON
    mode: String!
    contentItem: ContentItem
  }

  type ContentItem {
    id: ID
    name: String
    config: JSON
    sessions: [ContentItemSession]
  }

  type ItemSession {
    id: ID
  }

  enum SessionEventType {
    load
    save
    score
    manualScore
  }

  type SessionScore {
    points: Int
    max: Int
    type: ScoreType
  }
  enum ScoreType {
    AUTO
    MANUAL
  }

  input SessionScoreInput {
    partialScoring: Boolean!
    points: Int!
    max: Int!
    type: ScoreType!
  }
  """
   one event on a session
  """
  type SessionEvent {
    timestamp: String!
    id: ID!
    sessionId: ID!
    contentItemId: ID!
    eventType: SessionEventType
    data: JSON
    score: JSON
    itemConfig: JSON
  }
  """
   Provide a Mock Session Event to Publish
  """
  input SessionEventInput {
    sessionId: ID!
    eventType: SessionEventType!
    score: JSON
    data: JSON
    itemConfig: JSON
    "If provided, logic to determine the response view be skipped and this string will be rendered instead"
    response: String
  }
  
  type SessionEventNew {
    item: JSON
    firstLoad: Boolean
    session: JSON
  }

  type Subscription {
    sessionEvents(sessions: [ID]!): SessionEvent
    watchSessions(ids: [ID]): SessionEventNew
  }

  type Mutation {
    startSession(sessionId: ID): ID!
    """
     Adds a mock SessionEvent for the given session id
    """
    addSessionEvent(sessionId: ID!, event: SessionEventInput!): ID!
    """
     Mocks a save event for an item type - sessionId should be in the format 'interactiontype-id' e.g. 'mc-1', 'mc-2'
    """
    mockSaveEvent(sessionId: ID!, response: JSON): ID!
    mockScoreEvent(sessionId: ID!, score: SessionScoreInput!): ID!
  }
`;
const parseIdType = (id) => {
    const parts = id === null || id === void 0 ? void 0 : id.split("-");
    parts.pop();
    const type = parts === null || parts === void 0 ? void 0 : parts.join("-");
    if (!type) {
        throw new Error("invalid id, for mock id use '$interactiontype-id' format");
    }
    return type;
};
const timestamp = () => {
    return { timestamp: new Date().toISOString() };
};
const resolvers = {
    JSON: GraphQLJSON,
    Query: {
        sessionEvent: () => "Resolved",
        contentItemSession: async (_parent, args) => {
            var type = parseIdType(args.id);
            const itemConfig = JSON.parse(fs.readFileSync(`${__dirname}/fixtures/${type}.json`, "utf-8"));
            return {
                id: args.id,
                mode: "mock",
                contentItem: {
                    config: itemConfig,
                },
            };
        },
    },
    Subscription: {
        sessionEvents: {
            subscribe: withFilter(() => pubsub.asyncIterator(SESSION_EVENT), (payload, variables) => {
                var _a, _b;
                if ((_a = variables === null || variables === void 0 ? void 0 : variables.sessions) === null || _a === void 0 ? void 0 : _a.includes((_b = payload === null || payload === void 0 ? void 0 : payload.sessionEvents) === null || _b === void 0 ? void 0 : _b.sessionId)) {
                    return true;
                }
            }),
        },
    },
    Mutation: {
        startSession: (_parent, args) => {
            return args.sessionId ? args.sessionId : shortId.generate();
        },
        addSessionEvent: (_parent, args) => {
            const id = shortId.generate();
            const { sessionId, event } = args;
            const pubEvent = Object.assign(Object.assign(Object.assign(Object.assign({}, timestamp()), event), { sessionId }), { id });
            pubsub.publish(SESSION_EVENT, { sessionEvents: pubEvent });
            console.log(`addSessionEventArgs ${JSON.stringify(args)}`);
            return "addSessionEvent";
        },
        mockSaveEvent: (_parent, args) => {
            const id = shortId.generate();
            const { sessionId, response } = args;
            var type = parseIdType(sessionId);
            if (response) {
                mockSaveEventData.data = [JSON.parse(response)];
            }
            else {
                mockSaveEventData.data = JSON.parse(fs.readFileSync(`${__dirname}/fixtures/${type}-response.json`, "utf-8"));
            }
            const event = { eventType: "save" };
            const pubEvent = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, timestamp()), event), mockSaveEventData), { sessionId }), { id });
            pubsub.publish(SESSION_EVENT, { sessionEvents: pubEvent });
            console.log(`mockSaveEvent ${JSON.stringify(args)}`);
            return "mockSaveEvent";
        },
        mockScoreEvent: (_parent, args) => {
            const id = shortId.generate();
            const event = { eventType: "score" };
            const { sessionId, score } = args;
            if (score.type === "MANUAL") {
                event.eventType = "manual-score";
            }
            const pubEvent = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, timestamp()), event), { score }), { sessionId }), { id });
            pubsub.publish(SESSION_EVENT, { sessionEvents: pubEvent });
            console.log(`scoreSessionEvent ${JSON.stringify(args)}`);
            return "scoreSessionEvent";
        },
    },
};
const schema = makeExecutableSchema({
    typeDefs,
    resolvers,
});
module.exports = new ApolloServer({
    schema,
    introspection: true,
    subscriptions: '/',
    // https://github.com/prisma-labs/graphql-playground/issues/320
    playground: { settings: { "schema.polling.enable": false } },
});
